package FmModel;

// FmModel -- this model is the fusion machine. It's just a simple
// non-distributed non-multithreaded implementation. It only works
// with pi calculus terms, and only with names. (no constants).
//
// As for the compositional translation from pi calculus into this
// machine (i.e. the Operators), this isn't so relevant. After all,
// the fusion machine is supposed to start with just a machine c:[P].
// However, because of the way I've designed this system, composition
// translation is the only way I have to build up a model. Therefore,
// a fusion machine state has two modes: 'nascent' (which contains
// just a single term) and 'normal' (which is the fusion machine proper).
// 
// As for the compositional translation from the machine into the
// calculus, that's easy. Just like in the paper.
//

public class State extends Model.State
{ boolean isnascent; Term term; // If nascent, we just have a term.
  java.util.Dictionary chans; // a map from channelids to Channels.
  
  public State(Term at) {chans = new java.util.Hashtable(); isnascent=true; term=at;}
  
  public java.util.Vector EnumEqSteps()
  { GrowUp(); // ensure that we're no longer nascent
    java.util.Vector steps = new java.util.Vector();
    // we look for migrations of inputs, migration of outputs, deployments, from all channels
    java.util.Enumeration ci = chans.elements();
    while (ci.hasMoreElements())
    { Channel c = (Channel)ci.nextElement();
      if (c.fptr!=null) {for (int i=0; i<c.ins.size(); i++) steps.addElement(new Step(2,c.name,i,0));}
      if (c.fptr!=null) {for (int i=0; i<c.outs.size(); i++) steps.addElement(new Step(3,c.name,i,0));}
      for (int i=0; i<c.dep.size(); i++) steps.addElement(new Step(4,c.name,i,0));
    }
    return steps;
  }
  public java.util.Vector EnumExSteps()
  { GrowUp(); // ensure that we're no longer nascent
    // we look for interaction of elements
    java.util.Vector steps = new java.util.Vector();
    java.util.Enumeration ci = chans.elements();
    while (ci.hasMoreElements())
    { Channel c = (Channel)ci.nextElement();
      for (int i=0; i<c.ins.size(); i++) // which input atom
      { for (int j=0; j<c.outs.size(); j++) // which output atom
        { steps.addElement(new Step(1,c.name,i,j));
        }
      }
    }
    return steps;
  }
  public void Step(Object o) throws Exception
  { Step step = (Step)o;
    Channel c = (Channel)chans.get(step.cid);
    if (step.type==1)
    { // a reaction within channel c between input atom n and output atom n2
      Atom ain = c.ins.atomAt(step.n), aout=c.outs.atomAt(step.n2);
      Term P=ain.P, Q=aout.P;
      // If the atoms were replicated, then we'll leave them in place and
      // put (perhaps-substitued) term copies into the deployment area. If not,
      // we'll substitute their terms directly.
      if (ain.isrep) P=P.cloneTerm(); else c.ins.removeElementAt(step.n);
      if (aout.isrep) Q=Q.cloneTerm(); else c.outs.removeElementAt(step.n2);
      if (ain.dat.size()!=aout.dat.size()) throw new Exception("Arities don't match");
      if (ain.isbound && !aout.isbound) P.subst(aout.dat,ain.dat);
      else if (!ain.isbound && aout.isbound) Q.subst(ain.dat,aout.dat);
      else if (!ain.isbound && !aout.isbound) {for (int i=0; i<ain.dat.size(); i++) c.dep.addElement(new FusTerm(ain.dat.nocidAt(i),aout.dat.nocidAt(i)));}
      else
      { // doubly-bound reaction requires us to create fresh channels. We'll work
        // from right-to-left, using the 'out' names as stems for the new cids we create.
        for (int i=aout.dat.size()-1; i>=0; i--)
        { String nbase = Channelid.invertCase(aout.dat.nocidAt(i).n.n); int count=0;
          Channelid cid = new Channelid(nbase);
          while (chans.get(cid)!=null) {count++; cid.n = nbase+Integer.toString(count);}
          Channel sc = new Channel();
          sc.name = cid;
          sc.islocal = true;
          sc.fptr = null;
          chans.put(sc.name,sc);
          P.subst(cid,ain.dat.nocidAt(i));
          Q.subst(cid,aout.dat.nocidAt(i));
        }
      }
      c.dep.addElement(P); c.dep.addElement(Q);
    }
    else if (step.type==2 || step.type==3)
    { // A migration from c
      if (c.fptr==null) throw new Exception("Unable to migrate because no fusion pointer");
      Atom a; if (step.type==2) {a=c.ins.atomAt(step.n); c.ins.removeElementAt(step.n);}
      else {a=c.outs.atomAt(step.n); c.outs.removeElementAt(step.n);}
      Channel c2 = (Channel)chans.get(c.fptr);
      if (step.type==2) c2.ins.addElement(a);
      else c2.outs.addElement(a);
    }
    else if (step.type==4)
    { // a deployment of a term
      Term t = c.dep.termAt(step.n); c.dep.removeElementAt(step.n);
      // Following is my kudgy way of doing a 'typecase' in Java. Java really
      // wants me to use a virtual method, in an OO-sort of way. But I prefer
      // to use a function, in an ML-sort of way.
      ParTerm pt=null; try {pt=(ParTerm)t;} catch (Exception e) {}
      XmitTerm xt=null;try {xt=(XmitTerm)t;}catch (Exception e) {}
      NilTerm nt=null; try {nt=(NilTerm)t;} catch (Exception e) {}
      ResTerm rt=null; try {rt=(ResTerm)t;} catch (Exception e) {}
      FusTerm ft=null; try {ft=(FusTerm)t;} catch (Exception e) {}
      if (pt!=null) {c.dep.addElement(pt.P); c.dep.addElement(pt.Q);}
      else if (xt!=null)
      { Atom a = new Atom(xt.isrep,xt.dat,xt.isbound,xt.P);
        Channel c2 = (Channel)chans.get(xt.chan.cid);
        if (xt.issend) c2.outs.addElement(a);
        else c2.ins.addElement(a);
      }
      else if (nt!=null) {}
      else if (rt!=null)
      { String nbase = Channelid.invertCase(rt.r.n); int count=0;
        Channelid cid = new Channelid(nbase);
        while (chans.get(cid)!=null)
        { count++; cid.n = nbase+Integer.toString(count);
        }
        Channel sc = new Channel();
        sc.name = cid;
        sc.islocal = true;
        sc.fptr = null;
        chans.put(sc.name,sc);
        rt.P.subst(sc.name, rt.r);
        c.dep.addElement(rt.P);
      }
      else if (ft!=null)
      { int res = ft.n1.cid.compareTo(ft.n2.cid);
        Channelid f1,f2; // f1 is lesser and f2 is greater out of f1=f2
        if (res<0) {f1=ft.n1.cid; f2=ft.n2.cid;} else {f1=ft.n2.cid; f2=ft.n1.cid;}
        Channel c2 = (Channel)chans.get(f1);
        Channelid f3 = c2.fptr;
        // First, replace the target's fptr 
        c2.fptr = f2;
        // Now, perhaps, add a fusion into the deployment area
        if (f3!=null) c2.dep.addElement(new FusTerm(f1,f3));
      }
    }
  }
  
  
  public String toString()
  { if (isnascent) return "src: "+term.toString();
    java.util.Enumeration i = chans.elements();
    String s=""; boolean isfirst=true;
    while (i.hasMoreElements())
    { Channel chan = (Channel)i.nextElement();
      if (!isfirst) s+="  "; isfirst=false;
      s += chan.toString();
    }
    return s;
  }
  public Model.State Play(Model.Operators ops) throws Exception
  { if (isnascent) GrowUp(); // converts the nascent term into a chan.
    // First we translate the contents of each chan.
    Channelids res=new Channelids(); // meanwhile accumulating a list of all restricted names
    java.util.Enumeration i = chans.elements();
    Model.State s=null;
    while (i.hasMoreElements())
    { Channel chan = (Channel)i.nextElement();
      if (chan.islocal) res.addElement(chan.name);
      if (chan.fptr!=null) s=ops.Fus(chan.name.n,chan.fptr.n);
      java.util.Enumeration j=chan.ins.elements(); 
      while (j.hasMoreElements())
      { Model.State s2 = ((Atom)j.nextElement()).Play(ops,chan.name,false);
        if (s==null) s=s2; else s=ops.Par(s,s2);
      }
      j = chan.outs.elements();
      while (j.hasMoreElements())
      { Model.State s2 = ((Atom)j.nextElement()).Play(ops,chan.name,true);
        if (s==null) s=s2; else s=ops.Par(s,s2);
      }
      j=chan.dep.elements();
      while (j.hasMoreElements())
      { Model.State s2 = ((Term)j.nextElement()).Play(ops);
        if (s==null) s=s2; else s=ops.Par(s,s2);
      }
    }
    // Then we apply global restrictions
    if (s==null) s=ops.Nil();
    if (res.size()==0) return s;
    String[] rn = new String[res.size()];
    for (int j=0; j<res.size(); j++) rn[j]=res.cidAt(j).n;
    return ops.Res(rn,s);  
  }
  
  void GrowUp()
  { if (!isnascent) return;
    isnascent=false;
    // First we create empty channels for all the free names listed there
    Names fns = term.fns();
    for (int i=0; i<fns.size(); i++)
    { Channel sc = new Channel();
      sc.name = new Channelid(Channelid.invertCase(fns.nameAt(i).n));
      sc.islocal = false;
      sc.fptr = null;
      chans.put(sc.name,sc);
      term.subst(sc.name,fns.nameAt(i));      
    }
    // Now we create the 'src:' channel which contains the original source code
    Channel c = new Channel();
    c.name = new Channelid("src");
    c.islocal=false;
    c.fptr=null;
    c.dep.addElement(term);
    chans.put(c.name,c);
    term=null;
  }
}

class Step
{ int type;
  Channelid cid;
  int n,n2;
  // type=1: a reaction within channel cid, between input atom n and output atom n2
  // type=2: a migration of input atom n from channel cid
  // type=3: a migration of output atom n from channel cid
  // type=4: a deployment of term n from channel cid
  Step(int atype,Channelid acid,int an,int an2) {type=atype; cid=acid; n=an; n2=an2;}
  public String toString()
  { if (type==1) return "Reaction at "+cid.toString()+" in#"+Integer.toString(n)+" out#"+Integer.toString(n2);
    else if (type==2) return "Migration from "+cid.toString()+" of in#"+Integer.toString(n);
    else if (type==3) return "Migration from "+cid.toString()+" of out#"+Integer.toString(n);
    else return "Deployment from "+cid.toString()+" of dep#"+Integer.toString(n);
  }
} 


// At runtime, 'channelids' are the opaque names by which channels
// are identified.
class Channelid
{ String n;
  Channelid(String an) {n=an;}
  public int hashCode() {return n.hashCode();}
  public boolean equals(Channelid a) {return n.equals(a.n);}
  public boolean equals(Object o) {try {Channelid co=(Channelid)o; return n.equals(co.n);} catch (Exception e) {return false;}}
  // nb. the hashtable and other contains use this generic equals(object) method. It's only
  // our own code that will use the specific equals(channelid) one.
  public int compareTo(Channelid a) {return n.compareTo(a.n);}
  public int compareTo(Object o) {try {Channelid co=(Channelid)o; return n.compareTo(co.n);} catch (Exception e) {return 1;}}
  public String toString() {return n;}
  public static String invertCase(String s)
  { char[] c = s.toCharArray();
    for (int i=0; i<c.length; i++)
    { if (Character.isUpperCase(c[i])) c[i]=Character.toLowerCase(c[i]);
      else if (Character.isLowerCase(c[i])) c[i]=Character.toUpperCase(c[i]);
    }
    return new String(c);
  }
}

// 'names' are the textual names used in pieces of source code.
class Name
{ String n;
  Name(String an) {n=an;}
  public boolean equals(Name a) {return n.equals(a.n);}
  public boolean equals(Object o) {try {Name ao=(Name)o; return n.equals(ao.n);} catch (Exception e) {return false;}}
  // again, vector and other contains will use this generic equals(object) version
  public String toString() {return n;}
}

class Channel
{ Channelid name; boolean islocal; // islocal means (|n|).
  Channelid fptr;
  Atoms ins, outs;
  Terms dep;
  Channel() {ins=new Atoms(); outs=new Atoms(); dep=new Terms(); fptr=null;}
  public String toString()
  { String head="",tail="";
    if (islocal) {head="("; tail=")";}
    String s = head+name.toString();
    if (fptr!=null) s+="_"+fptr.toString();
    s+=tail+":[";
    boolean jfirst=true;
    java.util.Enumeration j=ins.elements(); 
    while (j.hasMoreElements())
    { if (!jfirst) s+=";"; jfirst=false;
      Atom a = (Atom)j.nextElement(); s+=a.toAtomString("in");
    }
    j=outs.elements();
    while (j.hasMoreElements())
    { if (!jfirst) s+=";"; jfirst=false;
      Atom a = (Atom)j.nextElement(); s+=a.toAtomString("out");
    }
    j=dep.elements();
    while (j.hasMoreElements())
    { if (!jfirst) s+=";"; jfirst=false;
      Term t = (Term)j.nextElement();
      s+=t.toString();
    }
    s+="]";
    return s;
  }
}

class Atom
{ boolean isrep;
  NamesOrChannelids dat; boolean isbound;
  Term P;
  Atom(boolean aisrep,NamesOrChannelids adat,boolean aisbound,Term aP) {isrep=aisrep; dat=adat; isbound=aisbound; P=aP;}
  public String toAtomString(String type)
  { String s="";
    if (isrep) s+="!";
    s+=type;
    if (isbound) s+="(";
    s+=dat.toStringList();
    if (isbound) s+=")";
    s+="."+P.toString();
    return s; 
  }
  public Model.State Play(Model.Operators ops,Channelid cid,boolean issend) throws Exception
  { Model.State s = P.Play(ops);
    String[] sdat = new String[dat.size()];
    for (int i=0; i<dat.size(); i++) sdat[i]=dat.nocidAt(i).toString();
    if (issend) s=ops.Snd(cid.n,sdat,isbound,s);
    else s=ops.Rcv(cid.n,sdat,isbound,s);
    if (isrep) s=ops.Rep(s);
    return s;
  }
} 


abstract class Term
{ public String toString() {return "(Term)";}
  public Names fns() {return new Names();}
  public abstract Model.State Play(Model.Operators ops) throws Exception;
  public abstract Term cloneTerm();
  public abstract void subst(Channelid snew,NameOrChannelid sold); 
  // subst is a destructive (in-place) substitution. NB. restriction
  // is an operation on names, not channelids, so our substitution is
  // easier than normal capture-avoiding substitution.
  public void subst(Channelid snew,Name sold) {subst(snew,new NameOrChannelid(sold));}
  public void subst(NamesOrChannelids snews, NamesOrChannelids solds)
  { for (int i=0; i<snews.size(); i++)
    { Channelid snew = snews.nocidAt(i).cid; // nb. we assume that all snews are cids.
      subst(snew,new NameOrChannelid(solds.nocidAt(i)));
    }
  }
}
class ParTerm extends Term
{ Term P, Q;
  ParTerm(Term aP,Term aQ) {P=aP; Q=aQ;}
  public void subst(Channelid snew,NameOrChannelid sold) {P.subst(snew,sold); Q.subst(snew,sold);}
  public String toString() {return "("+P.toString()+"|"+Q.toString()+")";}
  public Names fns() {Names fns = P.fns(); fns.union(Q.fns()); return fns;}
  public Model.State Play(Model.Operators ops) throws Exception {return ops.Par(P.Play(ops),Q.Play(ops));}
  public Term cloneTerm() {return new ParTerm(P.cloneTerm(),(Term)Q.cloneTerm());}
}
class XmitTerm extends Term
{ boolean isrep;
  NameOrChannelid chan; boolean issend;
  NamesOrChannelids dat; boolean isbound;
  Term P;
  XmitTerm() {dat=new NamesOrChannelids();} // this one's too big to construct from arguments to the constructor.
  public Term cloneTerm() {XmitTerm x=new XmitTerm(); x.isrep=isrep; x.chan=new NameOrChannelid(chan); x.issend=issend; x.dat=dat.cloneNamesOrChannelids(); x.isbound=isbound; x.P=P.cloneTerm(); return x;}
  public void subst(Channelid snew,NameOrChannelid sold)
  { chan.subst(snew,sold);
    if (isbound && dat.contains(sold)) return;
    if (!isbound) dat.subst(snew,sold);
    P.subst(snew,sold);
  }
  public Model.State Play(Model.Operators ops) throws Exception
  { Model.State s = P.Play(ops);
    String[] sdat = new String[dat.size()];
    for (int i=0; i<dat.size(); i++) sdat[i]=dat.nocidAt(i).toString();
    if (issend) s=ops.Snd(chan.toString(),sdat,isbound,s);
    else s=ops.Rcv(chan.toString(),sdat,isbound,s);
    if (isrep) s=ops.Rep(s);
    return s;   
  }
  public Names fns()
  { Names fns = P.fns();
    if (isbound)
    { for (int i=0; i<dat.size(); i++)
      { NameOrChannelid nocid = dat.nocidAt(i);
        if (nocid.isname) fns.removeElement(nocid.n);
      }
    }
    else
    { for (int i=0; i<dat.size(); i++)
      { NameOrChannelid nocid = dat.nocidAt(i);
        if (nocid.isname) fns.addElement(nocid.n);
      }
    }
    if (chan.isname) fns.addElement(chan.n);
    return fns;
  }
  public String toString()
  { String s="";
    if (isrep) s+="!";
    s+=chan.toString();
    String head="",tail="";
    if (issend && isbound) {head="!("; tail=")";}
    else if (issend && !isbound) {head="<"; tail=">";}
    else if (!issend && isbound) {head="("; tail=")";}
    else if (!issend && !isbound) {head="?"; tail="";}
    if (dat.size()==0) {tail=""; if (issend) head="!"; else head="?";}
    s+=head+dat.toStringList()+tail;
    try {NilTerm nt = (NilTerm)P;}
    catch (Exception e) {s+="."+P.toString();}
    return s;
  }
}

class NilTerm extends Term
{ public void subst(Channelid snew,NameOrChannelid sold) {}
  public String toString() {return "0";}
  public Names fns() {return new Names();}
  public Model.State Play(Model.Operators ops) throws Exception {return ops.Nil();}
  public Term cloneTerm() {return new NilTerm();}
}
class ResTerm extends Term
{ Name r;
  Term P;
  ResTerm(Name ar,Term aP) {r=ar; P=aP;} 
  public void subst(Channelid snew,NameOrChannelid sold) {if (sold.equals(r)) return; P.subst(snew,sold);}
  public String toString() {return "("+r.toString()+")("+P.toString()+")";}
  public Names fns() {Names fns=P.fns(); fns.removeElement(r); return fns;}
  public Model.State Play(Model.Operators ops) throws Exception
  { String[] rs = new String[1]; rs[0]=r.n;
    Model.State op = P.Play(ops);
    return ops.Res(rs,op);
  }
  public Term cloneTerm() {return new ResTerm(r,P.cloneTerm());}
}  
class FusTerm extends Term
{ NameOrChannelid n1,n2;
  FusTerm(NameOrChannelid an1,NameOrChannelid an2) {n1=an1; n2=an2;}
  FusTerm(Name an1,Name an2) {n1=new NameOrChannelid(an1); n2=new NameOrChannelid(an2);}
  FusTerm(Channelid an1,Channelid an2) {n1=new NameOrChannelid(an1); n2=new NameOrChannelid(an2);}
  public void subst(Channelid snew,NameOrChannelid sold) {n1.subst(snew,sold); n2.subst(snew,sold);}
  public String toString() {return n1.toString()+"="+n2.toString();}
  public Names fns() {Names fns = new Names(); if (n1.isname) fns.addElement(n1.n); if (n2.isname) fns.addElement(n2.n); return fns;}
  public Model.State Play(Model.Operators ops) throws Exception {return ops.Fus(n1.toString(),n2.toString());}
  public Term cloneTerm() {return new FusTerm(new NameOrChannelid(n1), new NameOrChannelid(n2));}
}





// These are all container classes. I write them out explicitly like
// this, so that the types of my container classes are informative.
class Atoms extends java.util.Vector
{ public final synchronized Atom atomAt(int i) {return (Atom)elementAt(i);}
}
class Terms extends java.util.Vector
{ public final synchronized Term termAt(int i) {return (Term)elementAt(i);}
}
class NameOrChannelid
{ boolean isname; Name n; Channelid cid;
  NameOrChannelid(boolean aisname,String s) {isname=aisname; if (isname) n=new Name(s); else cid=new Channelid(s);} 
  NameOrChannelid(Name an) {isname=true; n=an;}
  NameOrChannelid(Channelid acid) {isname=false; cid=acid;}
  NameOrChannelid(NameOrChannelid a) {isname=a.isname; n=a.n; cid=a.cid;}
  public boolean equals(NameOrChannelid a) {if (isname && a.isname) return n.equals(a.n); else if (!isname && !a.isname) return cid.equals(a.cid); else return false;}
  public boolean equals(Name a) {if (isname) return n.equals(a); else return false;}
  public boolean equals(Channelid a) {if (isname) return false; else return cid.equals(a);}
  public boolean equals(Object a) {NameOrChannelid ncid=(NameOrChannelid)a; if (ncid!=null) return equals(ncid);  Name an=(Name)a; if (an!=null) return equals(an); Channelid acid=(Channelid)a; if (a!=null) return equals(acid); return false;}
  public void subst(Channelid snew, NameOrChannelid sold)
  { if (this.equals(sold))
    { isname=false; cid=snew;
    }
  }
  public String toString() {if (isname) return n.toString(); else return cid.toString();}
}
class NamesOrChannelids extends java.util.Vector
{ public final synchronized NameOrChannelid nocidAt(int i) {return (NameOrChannelid)elementAt(i);}
  public void subst(Channelid snew, NameOrChannelid sold)
  { for (int i=0; i<size(); i++)
    { Object o = elementAt(i);
      NameOrChannelid nocid = nocidAt(i);
      nocid.subst(snew,sold);
    }
  }
  public String toStringList() {String s=""; for (int i=0; i<size(); i++) {if (i>0)s+=","; s+=nocidAt(i);} return s;}
  public NamesOrChannelids cloneNamesOrChannelids()
  { NamesOrChannelids n = new NamesOrChannelids();
    for (int i=0; i<size(); i++) n.addElement(new NameOrChannelid(nocidAt(i)));
    return n;
  }
}

class Channelids extends java.util.Vector
{ public final synchronized Channelid cidAt(int i) {return (Channelid)elementAt(i);}
}

class Names extends java.util.Vector
{ public final synchronized Name nameAt(int i) {return (Name)elementAt(i);}
  public final synchronized void union(Names b)
  { for (int i=0; i<b.size(); i++)
    { Name bn = b.nameAt(i);
      if (!this.contains(bn)) this.addElement(bn);
    }
  }
}