package TextModel;
import java.io.*;

// This is a model designed to translate into text (i.e. print stuff out)
// and translate out of text (i.e. parse a text string).
//
// When the state (i.e. the text string) is set explicitly,
// this contains the definitive state. It can be played (i.e.
// parsed) at will.
//
// But when the state is being constructed through the use of
// operators, it means that someone is using us to print
// stuff out. In this case, the text string remains empty,
// and we instead store the state in our own custom way. This
// is so we can perform optimisations before outputting text --
// for instance, to avoid redundant parentheses. We only turn
// it into text at the last minute, when needed. The value of
// the state is (bounds)(pars)


public class State extends Model.State
{ static AutoGram ag=null; // we cache our parser.
  String s; // this is the state, as a string
  // Or, maybe s is empty, and instead the state is stored in
  // the following, denoting (res)(sb). Ispar/issum might
  // indicate that sb is an unparenthesised sum or par. (If neither,
  // then it's atomic, so we can write (res)sb on its own.)
  String sb; boolean ispar; boolean issum;
  Names res;

  public State() {s=""; sb=""; ispar=false; issum=false; res=new Names();}  
  public State(String as) {s=as; sb=""; ispar=false; issum=false; res=new Names();}
  public void setString(String as) {s=as;}
  // toString: either the state is being stored as a string (in
  // which case we can return it directly) or it is being
  // stored in its alternative form (in which case we have to
  // construct it).
  public String toString()
  { if (s!="") return s;
    return Collapse(true,true);
  }
   
    
  // Collapse(parokay,sumsokay). This turns the current state
  // (stored in the fields above) into a string. The string
  // is protected with parentheses. Except: if it's a primitive
  // (i.e. fusion or send/receive or nil) then it doesn't need
  // parentheses, and if it's a PAR and parokay then it doesn't
  // need them, and if it's a SUM and sumokay then it doesn't
  // need them.
  public String Collapse(boolean parokay, boolean sumokay)
  { if (s!="") return "***"+s+"***";
    String a=sb;
    if (sb.length()==0) {a="0"; ispar=false; issum=false;}
    if (res.size()>0)
    { if (ispar || issum) a="("+a+")";
      return "("+res.toStringList()+")"+a;
    }
    else
    { if (ispar && parokay) return a;
      if (issum && sumokay) return a;
      if (!ispar || !issum) return a;
      return "("+a+")";
    }
  }


  // Play: this converts the current state from a string into
  // the pi/fusion calculus. It uses the grammer defined in gram.jj
  public Model.State Play(Model.Operators ops) throws Exception
  { if (s.equals("")) s=toString(); // might have to convert fields into string.
    InputStream stream = new ByteArrayInputStream(s.getBytes());
    if (ag==null) ag=new AutoGram(stream); else ag.ReInit(stream);
    ag.SetModel(ops);
    Model.State e=ag.All();
    return e;
  }

}




// Names: just a simple class to contain a list of names. And also
// to turn this list into a comma-delimited string.
class Names extends java.util.Vector
{ public final synchronized String nameAt(int i)
  { return (String)elementAt(i);
  }
  public String toStringList()
  { String s="";
    for (int i=0; i<size(); i++) {if (i!=0) s=s+","; s=s+nameAt(i);}
    return s;
  }

}
