package Ptrs;

public class Operators extends Model.Operators
{ 


  // Nil() -- ,,{}
  public Model.State Nil() {return new State();}

  // Dat(d) -- "d"=n, <n>, {}
  public Model.State Dat(String d)
  { State s=new State(); Name n=new Name(); n.title=d; s.names.addElement(n);
    s.frees=new String[1]; s.frees[0]=d; s.freen=new Name[1]; s.freen[0]=n; n.nref++;
    s.dats=new Name[1]; s.dats[0]=n; n.nref++;
    return s;
  }

  // Fusion(a,b) -- "a"=n "b"=m, <>, {n->m}
  public Model.State Fus(String a,String b) throws Exception
  { if (a.equals(b)) return Nil();
    State s=new State();
    Name n=new Name(); n.title=a; s.names.addElement(n);
    Name m=new Name(); m.title=b; s.names.addElement(m);
    s.frees=new String[2]; s.freen=new Name[2];
    s.frees[0]=a; s.freen[0]=n; n.nref++; s.frees[1]=b; s.freen[1]=m; m.nref++;
    if (n.id<m.id) n.fuseto(m); else m.fuseto(n);
    return s;
  }

  // Par(sa,sb) -- , <sadat,sbdat>, {sa + sb}
  // If sa and sb shared any free names, then the resulting state treats them as fusions
  public Model.State Par(Model.State a,Model.State b) throws Exception
  { State sa=(State)a, sb=(State)b; State s=new State();
    // Although we create a new State, we merely do a shallow copy of all the names.
    // That means that all the xmits still point to each other, and to names, as they
    // did before. We are not creating any new names.
    s.addStore(sa); s.addStore(sb);
    // We copy the old dats lists into our new one. That way reference counts stay the same
    s.dats=new Name[sa.dats.length+sb.dats.length];
    for (int i=0; i<sa.dats.length; i++) s.dats[i]=sa.dats[i];
    for (int i=0; i<sb.dats.length; i++) s.dats[sa.dats.length+i]=sb.dats[i];
    // Every free name either remains a free name, or becomes a fusion.
    MergeFrees(s,sa,sb);
    // and that's it!
    return s; 
  }


  
  // App(sa,sb) -- , , {sa + sb with sadat=sbdat}
  // If sa and sb shared any free names, then the resulting state treats them as fusions
  // (in addition to all the fusions created from equating datums)
  public Model.State App(Model.State a,Model.State b) throws Exception
  { State sa=(State)a, sb=(State)b; State s=new State();
    if (sa.dats.length != sb.dats.length) throw new Exception("Tried to APP different dat names");
    // Once again, a shallow copy of names
    s.addStore(sa); s.addStore(sb); 
    // Now turn each pair of dats into a fusion. (And remove the reference count from the dat before)
    for (int i=0; i<sa.dats.length; i++)
    { Name da=sa.dats[i], db=sb.dats[i];
      if (da.id<db.id) da.fuseto(db); else db.fuseto(da);
      da.nref--; db.nref--;
    }
    // Now we handle free names exactly as we did for PAR.
    MergeFrees(s,sa,sb);
    return s;
  }


  
  public void MergeFrees(State s,State sa,State sb) throws Exception
  { FStrings fas=new FStrings(sa.frees); FNames fan=new FNames(sa.freen);
    FStrings fbs=new FStrings(sb.frees); FNames fbn=new FNames(sb.freen);
    FStrings fs=new FStrings(); FNames fn1=new FNames(), fn2=new FNames();
    // We'll put string names in the array "fs"
    // fn1 will contain the primary thing that it will point to
    // fn2 will contain the fusion that might have to be introduced
    // This first pass will temporarily remove all free names, decrementing nrefs appropriately.
    // They will be added later.
    for (int i=0; i<fas.size(); i++)
    { if (fbs.containsString(fas.stringAt(i))) // two shared names!
      { int ib=fbs.indexOfString(fas.stringAt(i));
        Name n1=fan.nameAt(i), n2=fbn.nameAt(ib);
        n1.nref--; n2.nref--;
        if (n1.id<n2.id) {fn1.addElement(n1); fn2.addElement(n2);}
        else {fn1.addElement(n2); fn2.addElement(n1);}
        fs.addElement(fas.stringAt(i));
        fbs.setElementAt("",ib); // yucky hack so that the other guy doesn't try to add it again
      }
      else
      { fan.nameAt(i).nref--;
        fn1.addElement(fan.nameAt(i)); fn2.addElement(fan.nameAt(i)); // a duplicate in both arrays
        fs.addElement(fas.stringAt(i));
      }
    }
    for (int i=0; i<fbs.size(); i++) // these are by assumption not shared
    { if (fbs.stringAt(i).length()>0)
      { fbn.nameAt(i).nref--;
        fn1.addElement(fbn.nameAt(i)); fn2.addElement(fbn.nameAt(i));
        fs.addElement(fbs.stringAt(i));
      }
    }
    //
    // Second pass. Time now to increment nrefs appropriately.
    s.frees=new String[fs.size()]; s.freen=new Name[fn1.size()];
    for (int i=0; i<fs.size(); i++)
    { s.frees[i]=fs.stringAt(i); s.freen[i]=fn2.nameAt(i); // point to the greater
      s.freen[i].nref++;
    }
    //
    for (int i=0; i<fs.size(); i++)
    { Name n1=fn1.nameAt(i), n2=fn2.nameAt(i);
      if (n1!=n2) n1.fuseto(n2);
    }
  }


  // Res(r,s) -- \"r", , {s}
  // Just hide the free-names. If we restrict anything that wasn't free, that doesn't
  // actually do anything
  public Model.State Res(String[] rs,Model.State ss)
  { State s=(State)ss;
    int nsub=0; // this is how many we end up actually restricting
    // Now perform a first pass through and turn anything restricted into ""
    for (int i=0; i<rs.length; i++)
    { String r=rs[i];
      for (int j=0; j<s.frees.length; j++)
      { if (s.frees[j].equals(r)) {s.frees[j]=""; s.freen[j].nref--; s.freen[j]=null; nsub++;}
      }
    }
    // Now a copy pass through to copy all those free names remaining into a fresh array
    String[] olds=s.frees; Name[] oldn=s.freen; 
    s.frees=new String[s.frees.length-nsub]; s.freen=new Name[s.freen.length-nsub]; int ns=0;
    for (int i=0; i<olds.length; i++)
    { if (olds[i].length() != 0) {s.frees[ns]=olds[i]; s.freen[ns]=oldn[i]; ns++;}
    } 
    return s;
  }
  
  // Rep(s) -- , , {!s}
  // Consider ! (nu xyz)(v=z | y | x.z)  ==  (nu yz) <y> | v=z | !(nu x)(x.z)
  // The names that get replicated are only those not in the datum/freename interface.
  // All xmits, sums, reps get replicated.
  // But anything already covered by an existing replication does not get replicated.
  public Model.State Rep(Model.State ss)
  { State s=(State)ss; Rep r=new Rep();
    // First, say that we're going to replicate all thoroughly bound names
    // As the example above showed, a name doesn't get created by a replication if it was in the
    // interface or if it was fused to a free name. So, from the entire set of names, keep
    // on removing things linked to by the free names or linking to free names, until we're left
    // with a clique.
    Names free=new Names(), trapped=new Names();
    // Assume all names will be replicated (trapped). Then remove those that are free or datums. (free)
    // Then remove anything linked to by a free name.
    for (int i=0; i<s.names.size(); i++) trapped.addElement(s.names.nameAt(i));
    for (int i=0; i<s.dats.length; i++) {free.addElement(s.dats[i]); trapped.removeElement(s.dats[i]);}
    for (int i=0; i<s.freen.length; i++) {free.addElement(s.freen[i]); trapped.removeElement(s.freen[i]);}
    boolean anyfreed=true; while (anyfreed) // It doesn't matter if the size of the arrays change while we're iterating through them
    { anyfreed=false;
      for (int i=0; i<free.size(); i++) // we'll remove anything pointed to by a free
      { Name fpto=free.nameAt(i).fusion; if (fpto!=null && trapped.contains(fpto)) {free.addElement(fpto); trapped.removeElement(fpto); anyfreed=true;}
      }
      for (int i=0; i<trapped.size(); i++) // we'll remove anything that points to a free
      { Name tpto=trapped.nameAt(i).fusion; if (tpto!=null && free.contains(tpto)) {free.addElement(tpto); trapped.removeElement(tpto); anyfreed=true;}
      }
    }
    // Now, 'trapped' has all the names that will be genuinely replicated by the replication
    r.res=new Name[trapped.size()];
    for (int i=0; i<trapped.size(); i++) {r.res[i]=trapped.nameAt(i); r.res[i].nref++;}
    // We need to identify every single xmit, rep, sum, and link this rep to it (and vice versa).
    // But: anything already parented by some rep doesn't get parented by this one as well
    // anything parented by a sum only needs the sum to be parented 
    Xmits xmits=new Xmits(); Reps reps=new Reps(); Sums sums=new Sums();
    for (int ni=0; ni<s.names.size(); ni++)
    { Name n=s.names.nameAt(ni);
      for (int xi=0; xi<n.snds.size(); xi++) xmits.addElement(n.snds.xmitAt(xi));
      for (int xi=0; xi<n.rcvs.size(); xi++) xmits.addElement(n.rcvs.xmitAt(xi));
      for (int xi=0; xi<n.snds_dormant.size(); xi++) xmits.addElement(n.snds_dormant.xmitAt(xi));
      for (int xi=0; xi<n.rcvs_dormant.size(); xi++) xmits.addElement(n.rcvs_dormant.xmitAt(xi));
    }
    for (int si=0; si<s.sums.size(); si++)
    { Sum sum=s.sums.sumAt(si); sums.addElement(sum);
      for (int xi=0; xi<sum.xmits.length; xi++) xmits.removeElement(sum.xmits[xi]);
    }
    for (int ri=0; ri<s.reps.size(); ri++)
    { Rep rep=s.reps.repAt(ri); reps.addElement(rep);
      for (int i=0; i<rep.xmits.length; i++) xmits.removeElement(rep.xmits[i]);
      for (int i=0; i<rep.sums.length; i++) sums.removeElement(rep.sums[i]);
      for (int i=0; i<rep.reps.length; i++) reps.removeElement(rep.reps[i]);
    }
    // Now the arrays xmits, reps, sums have every node that will be pointed to by us
    r.xmits=new Xmit[xmits.size()]; r.sums=new Sum[sums.size()]; r.reps=new Rep[reps.size()];
    for (int xi=0; xi<xmits.size(); xi++) {r.xmits[xi]=xmits.xmitAt(xi); r.xmits[xi].rep=r;}
    for (int si=0; si<sums.size(); si++)  {r.sums[si] =sums.sumAt(si);   r.sums[si].rep=r;}
    for (int ri=0; ri<reps.size(); ri++)  {r.reps[ri] =reps.repAt(ri);} // no parent-rep pointer for reps
    // tada! Finally done!
    s.reps.addElement(r);
    return s;
  }
  

  // Sum(a,b) -- , 0, {sa+sb}
  // Cannot sum if there are datums or fusions or replications or parallels at the top level.
  // Can only sum xmits or sums.
  // i.e. every single active node must be rooted to the same sum; or there must be only
  // a single active node. And no fusions.
  public Model.State Sum(Model.State ssa,Model.State ssb) throws Exception
  { State sa=(State)ssa, sb=(State)ssb;
    if (sa.dats.length>0 || sb.dats.length>0) throw new Exception("Cannot sum a thing with datums");
    boolean failed=false;
    for (int i=0; i<sa.names.size(); i++) if (sa.names.nameAt(i).fusion!=null) failed=true;
    for (int i=0; i<sb.names.size(); i++) if (sb.names.nameAt(i).fusion!=null) failed=true;
    if (failed) throw new Exception("Cannot sum something with fusions");
    Sum thesums[]=new Sum[2]; thesums[0]=null; thesums[1]=null;
    Xmit thexmits[]=new Xmit[2]; thexmits[0]=null; thexmits[1]=null; 
    for (int index=0; index<4; index++)
    { State ss; if (index<2) ss=sa; else ss=sb;
      Sum thesum=thesums[index/2]; Xmit thexmit=thexmits[index/2];
      for (int ni=0; ni<ss.names.size(); ni++)
      { Name n=ss.names.nameAt(ni);
        Xmits hang=n.snds; if (index==1 || index==3) hang=n.rcvs;
        for (int xi=0; xi<hang.size(); xi++)
        { Xmit x=hang.xmitAt(xi);
          Sum thissum=x.sum; Xmit thisxmit=x; if (thissum!=null) thisxmit=null;
          if (thesum==null && thexmit==null) {thesum=thissum; thexmit=thisxmit;}
          else if (thesum==null && thexmit!=null) {if (thissum!=thesum || thisxmit!=thexmit) failed=true;}
          else if (thesum!=null && thexmit==null) {if (thissum!=thesum || thisxmit!=thexmit) failed=true;}
          else failed=true;
        }
      }
      if (thesum!=null && thesum.rep!=null) failed=true;
      if (thexmit!=null && thexmit.rep!=null) failed=true;
      thesums[index/2]=thesum; thexmits[index/2]=thexmit;
    }
    if (thesums[0]==null && thexmits[0]==null) failed=true;
    if (thesums[1]==null && thexmits[1]==null) failed=true;
    if (failed) throw new Exception("Cannot sum something not an xmit or a sum");
    // So we're summing two things, both of which are okay
    // first we do a parallel composition on them (which perhaps merges free names)
    // (it doesn't matter if that ends up with fusions... that'll be like x=y | (P+Q)
    State s=(State)Par(sa,sb);
    // We have either thesum or thexmit on the left, and the same on the right, and have
    // to figure out how to merge them. There was no rep, so that's easy. Just xmits
    Sum sum=new Sum(); int n=0;
    if (thesums[0]==null) n=1; else n=thesums[0].xmits.length;
    if (thesums[1]==null) n=n+1; else n=n+thesums[1].xmits.length;
    sum.xmits=new Xmit[n]; n=0;
    if (thesums[0]==null) {sum.xmits[n]=thexmits[0]; n++;}
    else {for (int i=0; i<thesums[0].xmits.length; i++) {sum.xmits[n]=thesums[0].xmits[i]; n++;}}
    if (thesums[1]==null) {sum.xmits[n]=thexmits[1]; n++;}
    else {for (int i=0; i<thesums[1].xmits.length; i++) {sum.xmits[n]=thesums[1].xmits[i]; n++;}}
    // Now, our new sum points to all the appropriate xmits.
    s.sums.addElement(sum);
    if (thesums[0]!=null) s.sums.removeElement(thesums[0]);
    if (thesums[1]!=null) s.sums.removeElement(thesums[1]);
    return s;
  }  


  public Model.State Snd(String n,Model.State a) {return Xmit(n,(State)a,true);}
  public Model.State Rcv(String n,Model.State a) {return Xmit(n,(State)a,false);}

  // Xmit(n,s) -- 0, 0, {inactivate+n}
  // We inactivate everything. Then ensure that a free name "n" is present.
  // Then hang an atom off it.
  public Model.State Xmit(String xs,State s,boolean issend)
  { // First, let's collect up (and inactivate) all the currently-top-level-active atoms
    Xmits toenable = new Xmits();
    for (int ni=0; ni<s.names.size(); ni++)
    { Name n=s.names.nameAt(ni);
      for (int xi=0; xi<n.snds.size(); xi++) {Xmit x=n.snds.xmitAt(xi); toenable.addElement(x); n.snds.removeElement(x); n.snds_dormant.addElement(x);}
      for (int xi=0; xi<n.rcvs.size(); xi++) {Xmit x=n.rcvs.xmitAt(xi); toenable.addElement(x); n.rcvs.removeElement(x); n.rcvs_dormant.addElement(x);}
    }
    // Next, collect up the datums
    Name[] xdats=s.dats; s.dats=new Name[0]; // reference count is left as it was
    // Next, collect up the fusions
    Names xfu1=new Names(), xfu2=new Names();
    for (int ni=0; ni<s.names.size(); ni++)
    { Name n=s.names.nameAt(ni);
      if (n.fusion!=null)
      { xfu1.addElement(n); n.nref++;
        xfu2.addElement(n.fusion); // refcount is preserved -- we've removed the fusion link, and added another
        n.fusion=null;
      }
    }
    // Now, ensure that the free name "xs" is present
    int namefree=-1; Name xn;
    for (int i=0; i<s.frees.length; i++) if (s.frees[i].equals(xs)) namefree=i;
    if (namefree!=-1)
    { xn=s.freen[namefree];
    }
    else
    { xn=new Name(); xn.title=xs; s.names.addElement(xn); xn.nref++;
      String[] newfrees=new String[s.frees.length+1]; for (int i=0; i<s.frees.length; i++) newfrees[i]=s.frees[i]; newfrees[s.frees.length]=xs;
      Name[]   newfreen=new Name  [s.freen.length+1]; for (int i=0; i<s.freen.length; i++) newfreen[i]=s.freen[i]; newfreen[s.freen.length]=xn;
      s.frees=newfrees; s.freen=newfreen;
    }
    // Now create the atom!
    Xmit xmit=new Xmit();
    xmit.dats=xdats;
    xmit.fusions_src=new Name[xfu1.size()]; for (int i=0; i<xfu1.size(); i++) xmit.fusions_src[i]=xfu1.nameAt(i);
    xmit.fusions_dst=new Name[xfu2.size()]; for (int i=0; i<xfu2.size(); i++) xmit.fusions_dst[i]=xfu2.nameAt(i);
    xmit.sum=null;
    xmit.rep=null;
    xmit.toenable=toenable;
    if (issend) xn.snds.addElement(xmit); else xn.rcvs.addElement(xmit);
    // (remember: reference counting is not for the xmits, only for other things)
    // and that's it
    return s;
  }


}



class FStrings extends java.util.Vector
{ public final synchronized String stringAt(int i) {return (String)elementAt(i);}
  public final synchronized boolean containsString(String s)
  { for (int i=0; i<size(); i++) {if (s.equals(stringAt(i))) return true;}
    return false;
  }
  public final synchronized int indexOfString(String s)
  { for (int i=0; i<size(); i++) {if (s.equals(stringAt(i))) return i;}
    return -1;
  }
  public FStrings(String[] s) {for (int i=0; i<s.length; i++) addElement(s[i]);}
  public FStrings() {}
}
class FNames extends java.util.Vector
{ public final synchronized Name nameAt(int i) {return (Name)elementAt(i);}
  public FNames(Name[] n) {for (int i=0; i<n.length; i++) addElement(n[i]);}
  public FNames() {}
}